package biggests.gui;

import java.awt.*;
import java.awt.event.*;

import javax.swing.*;
import javax.swing.tree.*;

import biggests.tables.*;
import biggests.tree.*;
import biggests.utils.*;
import smadeira.biclustering.*;
import smadeira.ontologizer.*;

/**
 * <p>Title: Matrix Analysis Tables Panel</p>
 *
 * <p>Description: Displays tables of real values, tables of colors
 *                 or tables of symbols for matrices, biclusters or
 *                 groups of biclusters.</p>
 *
 * <p>Copyright:   Copyright (C) 2007  Joana Gonçalves
 *                 This program is free software; you can redistribute
 *                 it and/or modify it under the terms of the GNU General
 *                 Public License as published by the Free Software
 *                 Foundation; either version 3 of the License, or
 *                 any later version.
 *
 *                 This program is distributed in the hope that it will
 *                 be useful, but WITHOUT ANY WARRANTY; without even the
 *                 implied warranty of MERCHANTABILITY or FITNESS FOR A
 *                 PARTICULAR PURPOSE.  See the
 *                 <a href="http://www.gnu.org/licenses/gpl.html">
 *                 GNU General Public License</a> for more details.
 * </p>
 *
 * @author Joana P. Gonçalves
 */

public class PanelAnalyzingMatrixTables extends JPanel {
    /**
     * The type for values tables.
     */
    public static final int VALUES = 0;

    /**
     * The type for colors tables.
     */
    public static final int COLORS = 1;

    /**
     * The type for symbols tables.
     */
    public static final int SYMBOLS = 2;

    /**
     * Indicates the type of tables to create. Valid types are:<br/>
     *     <code>PanelAnalyzingMatrixTables.VALUES</code><br/>
     *     <code>PanelAnalyzingMatrixTables.COLORS</code><br/>
     *     <code>PanelAnalyzingMatrixTables.SYMBOLS</code><br/>
     */
    private int tableType;

    /**
     * <code>biclustersNumber</code> contains the number of biclusters tables
     * that the user may choose to see. If the panel holds tables of biclusters'
     * matrices, then these options will be included in a combo box.
     */
    private static final String[] biclustersNumber = {"first 5 biclusters",
            "first 10 biclusters",
            "first 20 biclusters",
            "first 50 biclusters",
            "first 100 biclusters",
            "all biclusters"};

    /**
     * The application's main frame.
     */
    protected BiggestsFrame mainFrame;

    /**
     * The array which contains the tables to be displayed
     * (may contain <code>TableValues</code>s, <code>TableColors</code>s
     * and <code>TableSymbols</code>s, depending on the <code>tableType</code>).
     */
    private TableValues[] tables;

    /**
     * The array with the scroll panes that hold the tables to be
     * displayed (for <code>VALUES</code> <code>tableType</code> only).
     */
    private JScrollPane[] scrolls;

    /**
     * The set of panels which holds the tables and their headers,
     * which will be displayed in this panel (for <code>COLORS</code>
     * and <code>SYMBOLS</code> <code>tableType</code>s only).
     */
    private JPanel[] panels;

    /**
     * This combo box will only be initialized if the tables to show
     * come from biclusters groups. If that happens, it is initialized with
     * <code>biclustersNumber</code> options.
     */
    private JComboBox biclustersNumberOption;

    /**
     * The layout of this panel.
     */
    private BoxLayout boxLayout;

    /**
     * The set of biclusters. This will be only initialized if the tables
     * to display come from bicluster matrices, i.e., if the object to display
     * is either a group of biclusters or a group of post-processed biclusters.
     */
    private Bicluster[] setOfBiclusters;

    /**
     * The node information object which contains the object (expression matrix,
     * preprocessed matrix, discretized matrix, bicluster, biclusters group,
     * post-processed bicluster or post-processed biclusters group) to
     * construct the tables from.
     */
    private NodeInfo nodeInfo;

    private TreePath selectionPath;

    /**
     * Creates a new panel to display tables. Supports three kinds of tables:
     * values' tables, colors' tables and symbols' tables.<br/> valid
     * <code>tableType</code>s are:<br/>
     * <code>PanelAnalyzingMatrixTables.VALUES</code><br/>
     * <code>PanelAnalyzingMatrixTables.COLORS</code><br/>
     * <code>PanelAnalyzingMatrixTables.SYMBOLS</code><br/>
     *
     * @param frame BiggestsFrame the application's main frame
     * @param tableType the <code>int</code> type of the tables to display
     *                  in this panel
     */
    public PanelAnalyzingMatrixTables(BiggestsFrame frame, int tableType) {
        mainFrame = frame;
        try {
            if ((tableType != VALUES) &&
                (tableType != COLORS) &&
                (tableType != SYMBOLS)) {
                throw new Exception("The type of the tables can only be:\n" +
                                    "PanelAnalyzingMatrixTables.VALUES;\n" +
                                    "PanelAnalyzingMatrixTable.COLORS;\n" +
                                    "PanelAnalyzingMatrixTables.SYMBOLS");
            }
        } catch (Exception ex) {
            ex.printStackTrace();
        }

        this.tableType = tableType;

        try {
            jbInit();
        } catch (Exception ex) {
            ex.printStackTrace();
        }
    }

    /**
     * Initializes this panel's components.
     *
     * @throws Exception
     */
    private void jbInit() throws Exception {
        boxLayout = new BoxLayout(this, BoxLayout.Y_AXIS);
        this.setLayout(boxLayout);
        this.setBorder(BorderFactory.createEmptyBorder(20, 20, 20, 20));
        setOfBiclusters = new Bicluster[0];
        tables = new TableValues[0];
        scrolls = new JScrollPane[0];
        panels = new JPanel[0];
    }

    public void paintAgain() {
        clear();
        this.paint(this.nodeInfo, this.selectionPath);
    }

    /**
     * Creates and displays the tables with data retrieved from the
     * node object associated with the given node info object.
     * <p>
     * If node info object contains a node object which has a single expression
     * matrix, then a single table is created with data from that matrix.
     * </p>
     * <p>
     * Otherwise, if the node info object contains a node object which is a
     * bicluster group or a post-processed bicluster group and, by that,
     * contains a set of biclusters (each one provided with its own matrix),
     * then several tables are created (one for each bicluster).
     * For efficiency reasons, by default, only the first 5 tables are
     * displayed (for the first 5 biclusters of the bicluster group). The user
     * is then asked to select another number of tables to be shown, if he/she
     * wants.
     * </p>
     *
     * @param nodeInfo <code>NodeInfo</code> the node information object,
     *                 which contains the node object that will provide
     *                 the data for the tables
     * @param selectionPath the selected <code>TreePath</code>
     */
    public void paint(NodeInfo nodeInfo, TreePath selectionPath) {
        if (nodeInfo == null) {
            return;
        }
        if (selectionPath == null) {
            return;
        }

        this.nodeInfo = nodeInfo;
        this.selectionPath = selectionPath;
        NodeObjectInterface nodeObject = nodeInfo.getNodeObject();
        int type = nodeInfo.getNodeObjectType();

        if ((type == NodeInfo.BICLUSTER_GROUP) ||
            (type == NodeInfo.POSTPROCESSED_GROUP)) {
            /**
             * the object is either a bicluster group or a post-processed group
             */

            /**
             * get the set of biclusters that form the bicluster group
             */
            setOfBiclusters = ((Biclustering)nodeObject).
                              getSetOfBiclusters();
            GOFrontEnd go = mainFrame.getPanelAnalyzingFunctionTable().
                            getGOFrontEndForSelectionPath(selectionPath);

            try {
                /**
                 * set the number of tables to be displayed (5 by default)
                 */
                int limit;
                if (setOfBiclusters.length >= 5) {
                    limit = 5;
                } else {
                    limit = setOfBiclusters.length;
                }

                /**
                 * create the array of tables to be displayed
                 */
                switch (tableType) {
                case VALUES:
                    tables = new TableValues[limit];
                    for (int i = 0; i < limit; i++) {
                        IMatrix matrix = (IMatrix) setOfBiclusters[i].
                                         getBiclusterSource().getMatrix();
                        if (matrix.hasMissingValues()) {
                            tables[i] = new TableValuesWithMissingValues(
                                    setOfBiclusters[i], go, mainFrame);
                        }
                        else {
                            tables[i] = new TableValues(setOfBiclusters[i],
                                    go, mainFrame);
                        }
                    }
                    break;
                case COLORS:
                    tables = new TableColors[limit];
                    for (int i = 0; i < limit; i++) {
                        IMatrix matrix = (IMatrix) setOfBiclusters[i].
                                         getBiclusterSource().getMatrix();
                        if (matrix.hasMissingValues()) {
                            tables[i] = new TableColorsWithMissingValues(
                                        setOfBiclusters[i], go, mainFrame);
                        }
                        else {
                            tables[i] = new TableColors(setOfBiclusters[i],
                                    go, mainFrame);
                        }
                    }
                    break;
                case SYMBOLS:
                    if (!(nodeObject instanceof CCC_Biclustering) &&
                        !(nodeObject instanceof E_CCC_Biclustering)) {
                        throw new InvalidNodeObjectTypeException(
                                "Tables of symbols can only be constructed from " +
                                "discretized matrices or biclusters/post-processed " +
                                "biclusters which contain discretized data");
                    }

                    tables = new TableSymbols[limit];
                    for (int i = 0; i < limit; i++) {
                        IMatrix matrix = (IMatrix) setOfBiclusters[i].
                                         getBiclusterSource().getMatrix();
                        if (matrix.hasMissingValues()) {
                            tables[i] = new TableSymbolsWithMissingValues(
                                        (CCC_Bicluster)setOfBiclusters[i], go, mainFrame);
                        }
                        else {
                            tables[i] = new TableSymbols(setOfBiclusters[i],
                                    go, mainFrame);
                        }
                    }
                    break;
                }

                /**
                 * set the combo box with biclusters' number options
                 */
                JPanel panel = new JPanel();
                BoxLayout bl = new BoxLayout(panel, BoxLayout.X_AXIS);
                panel.setLayout(bl);
                panel.setBorder(BorderFactory.createEmptyBorder(20, 20, 20, 20));
                biclustersNumberOption = new JComboBox(biclustersNumber);
                panel.add(biclustersNumberOption);
                for (int i = 0; i < 7; i++) {
                    panel.add(new JPanel());
                }

                /**
                 * add the item listener that will be listening to the changes
                 * in items' selection on the combo box
                 */
                biclustersNumberOption.addItemListener(new
                        PanelAnalyzingMatrixTables_biclustersNumberOption_itemAdapter(this));
                /**
                 * display the combo box
                 */
                this.add(panel);
            } catch (InvalidNodeObjectTypeException ex) {
                ex.printStackTrace();
            }
            catch (Exception ex) {
                ex.printStackTrace();
            }
        } else
        if ((((type == NodeInfo.ORIGINAL) || type == NodeInfo.PREPROCESSED) &&
            (((AbstractExpressionMatrix)nodeObject).hasMissingValues())) ||
            ((type == NodeInfo.DISCRETIZED) &&
            ((DiscretizedExpressionMatrix)nodeObject).hasMissingValues())) {
            GOFrontEnd go = mainFrame.getPanelAnalyzingFunctionTable().
                            getGOFrontEndForSelectionPath(selectionPath);
            try {
                if (tableType == COLORS) {
                    if (nodeInfo.getNodeObject() instanceof AbstractExpressionMatrix) {
                        tables = new TableColorsWithMissingValues[1];
                        tables[0] = new TableColorsWithMissingValues(
                                (AbstractExpressionMatrix)nodeInfo.getNodeObject(), go,
                                mainFrame);
                    }
                    else {
                        tables = new TableColorsWithMissingValues[1];
                        tables[0] = new TableColorsWithMissingValues(
                                (DiscretizedExpressionMatrix)nodeInfo.getNodeObject(), go,
                                mainFrame);
                    }
                } else
                if (tableType == VALUES) {
                    tables = new TableValuesWithMissingValues[1];
                    tables[0] = new TableValuesWithMissingValues(
                            (IMatrix)nodeInfo.getNodeObject(), go,
                            mainFrame);
                }
                else
                if (tableType == SYMBOLS &&
                    nodeObject instanceof DiscretizedExpressionMatrix) {
                    tables = new TableSymbolsWithMissingValues[1];
                    tables[0] = new TableSymbolsWithMissingValues(
                            (DiscretizedExpressionMatrix)nodeInfo.getNodeObject(), go,
                            mainFrame);
                }
            } catch (InvalidNodeObjectTypeException ex) {
                ex.printStackTrace();
            } catch (Exception ex) {
                ex.printStackTrace();
            }
        }
        else
        if (((nodeObject instanceof Bicluster) &&
             (((IMatrix)(((Bicluster)nodeObject).
                         getBiclusterSource().
                         getMatrix())).hasMissingValues()))) {
            GOFrontEnd go = mainFrame.getPanelAnalyzingFunctionTable().
                            getGOFrontEndForSelectionPath(selectionPath);
            try {
                if (tableType == COLORS) {
                        tables = new TableColorsWithMissingValues[1];
                        tables[0] = new TableColorsWithMissingValues(
                                (Bicluster)nodeInfo.getNodeObject(), go,
                                mainFrame);
                } else
                if (tableType == VALUES) {
                    tables = new TableValuesWithMissingValues[1];
                    tables[0] = new TableValuesWithMissingValues(
                            (Bicluster)nodeInfo.getNodeObject(), go,
                            mainFrame);
                }
                else
                if (tableType == SYMBOLS &&
                    nodeObject instanceof CCC_Bicluster) {
                    tables = new TableSymbolsWithMissingValues[1];
                    tables[0] = new TableSymbolsWithMissingValues(
                            (CCC_Bicluster)nodeInfo.getNodeObject(), go,
                            mainFrame);
                }
            } catch (InvalidNodeObjectTypeException ex) {
                ex.printStackTrace();
            } catch (Exception ex) {
                ex.printStackTrace();
            }
        }

        else {
            /**
             * the node object is either an original,
             * preprocessed, discretized matrix (without missing values)
             * or a single
             * bicluster/post-processed bicluster
             */
            try {
                /**
                 * set the array of tables to be displayed (only one table to
                 * display, in this case)
                 */
                GOFrontEnd go = mainFrame.getPanelAnalyzingFunctionTable().
                                getGOFrontEndForSelectionPath(selectionPath);
                switch (tableType) {
                case VALUES:
                    tables = new TableValues[1];
                    tables[0] = new TableValues(nodeObject, go, mainFrame);
                    break;
                case COLORS:
                    tables = new TableColors[1];
                    tables[0] = new TableColors(nodeObject, go, mainFrame);
                    break;
                case SYMBOLS:
                    if (!(nodeObject instanceof CCC_Bicluster) &&
                        !(nodeObject instanceof E_CCC_Bicluster) &&
                        !(nodeObject instanceof DiscretizedExpressionMatrix)) {
                        throw new InvalidNodeObjectTypeException(
                                "Tables of symbols can only be constructed from " +
                                "discretized matrices or biclusters/post-processed " +
                                "biclusters which contain discretized data");
                    }

                    tables = new TableSymbols[1];
                    tables[0] = new TableSymbols(nodeObject, go, mainFrame);
                    break;
                }
            } catch (InvalidNodeObjectTypeException ex) {
                ex.printStackTrace();
            }
        }
        this.displayTables();
    }

    /**
     * Clears the panel. Prepares it to receive new data for a new set of tables.
     * Removes all displayed components and initializes tables and scrolls arrays.
     */
    public void clear() {
        this.removeAll();
        this.boxLayout = new BoxLayout(this, BoxLayout.Y_AXIS);
        this.setLayout(boxLayout);
        setOfBiclusters = null;
//        selectionPath = null;
        tables = null;
        scrolls = null;
        panels = null;
    }

    /**
     *
     */
    private void clearPreserveBiclustersSet() {
        this.removeAll();
        tables = null;
        scrolls = null;
        panels = null;
    }

    /**
     * Displays the constructed tables (values, colors or symbols' tables).
     */
    private void displayTables() {
        if ((tableType == COLORS) || (tableType == SYMBOLS) ||
            ((setOfBiclusters != null) && (setOfBiclusters.length > 1) &&
             (tableType == VALUES))) {
            panels = new JPanel[tables.length];

            if (setOfBiclusters != null &&
                nodeInfo.getNodeObjectType() == NodeInfo.POSTPROCESSED_GROUP) {
                for (int i = 0; i < panels.length; i++) {
                    JPanel panel = new JPanel();
                    panel.setLayout(new BorderLayout());
                    tables[i].setBackground(this.getBackground());
                    tables[i].setBackground(this.getBackground());
                    panel.add(tables[i].getTableHeader(), BorderLayout.NORTH);
                    panel.add(tables[i], BorderLayout.WEST);
                    BorderLayout bl = new BorderLayout();
                    JLabel label = new JLabel("\t\tPost-Processed Bicluster "
                                              + (i + 1) + "   (" +
                                              setOfBiclusters[i].getID() + ")");
                    panels[i] = new JPanel(bl);
                    panels[i].setBorder(BorderFactory.createEmptyBorder(10, 10,
                            10,
                            10));
                    panels[i].add(label, BorderLayout.NORTH);
                    panels[i].add(panel, BorderLayout.WEST);
                }
            } else
            if (setOfBiclusters != null) {
                for (int i = 0; i < panels.length; i++) {
                    JPanel panel = new JPanel();
                    panel.setLayout(new BorderLayout());
                    tables[i].setBackground(this.getBackground());
                    tables[i].setBackground(this.getBackground());
                    panel.add(tables[i].getTableHeader(), BorderLayout.NORTH);
                    panel.add(tables[i], BorderLayout.WEST);
                    BorderLayout bl = new BorderLayout();
                    JLabel label = new JLabel("\t\tBicluster " +
                                              setOfBiclusters[i].getID());
                    panels[i] = new JPanel(bl);
                    panels[i].setBorder(BorderFactory.createEmptyBorder(10, 10,
                            10,
                            10));
                    panels[i].add(label, BorderLayout.NORTH);
                    panels[i].add(panel, BorderLayout.WEST);
                }
            } else {
                for (int i = 0; i < panels.length; i++) {
                    panels[i] = new JPanel();
                    panels[i].setBorder(BorderFactory.createEmptyBorder(10, 10,
                            10, 10));
                    panels[i].setLayout(new BorderLayout());
                    tables[i].setBackground(this.getBackground());
                    panels[i].add(tables[i].getTableHeader(),
                                  BorderLayout.NORTH);
                    panels[i].add(tables[i], BorderLayout.WEST);
                }
            }
            for (JPanel panel : panels) {
                this.add(panel);
            }
            return;
        }
        /**
         * set the set of scrolls that will scroll the created tables (one
         * scroll is needed for each table)
         */
        scrolls = new JScrollPane[tables.length];
        for (int i = 0; i < tables.length; i++) {
            scrolls[i] = new JScrollPane(tables[i]);
            scrolls[i].getVerticalScrollBar().
                setUnitIncrement(mainFrame.SCROLL_BAR_INCREMENT/2);
        }

        /**
         * display the set of scrolls, along with their tables
         */
        for (JScrollPane scrollTable : scrolls) {
            scrollTable.setBorder(BorderFactory.createEmptyBorder(10, 10, 10,
                    10));
            this.add(scrollTable);
        }
    }

    /**
     * Fired when the selected item of the biclusters' number combo box changes.
     * Checks the new number of tables to be displayed, creates and shows them.
     *
     * @param event the <code>ItemEvent</code> resulting of the selection of a
     *              new item in the <code>biclustersNumberOption</code> combo box
     */
    public void biclustersNumberOption_itemStateChanged(ItemEvent event) {
        if (event.getStateChange() == ItemEvent.DESELECTED) {
            return;
        }

        int limit;
        int selectedIndex = biclustersNumberOption.getSelectedIndex();

        /**
         * set the number of tables to be displayed
         */
        switch (selectedIndex) {
        case 0:
            limit = 5;
            break;
        case 1:
            limit = 10;
            break;
        case 2:
            limit = 20;
            break;
        case 3:
            limit = 50;
            break;
        case 4:
            limit = 100;
            break;
        case 5:
            limit = setOfBiclusters.length;
            break;
        default:
            limit = 5;
            break;
        }

        /**
         * if the number of biclusters to be displayed is higher than the
         * existing set of biclusters, then limit it to the number of biclusters
         * in the set
         */
        if (limit > setOfBiclusters.length) {
            limit = setOfBiclusters.length;
        }

        /**
         * clears this panel from its components
         */
        this.clearPreserveBiclustersSet();

        try {
            /**
             * set the tables to be displayed (one for each bicluster); the
             * number of tables is limited by the selection of the user (or by
             * the number of biclusters in the set)
             */
            GOFrontEnd go = mainFrame.getPanelAnalyzingFunctionTable().
                            getGOFrontEndForSelectionPath(selectionPath);
            switch (tableType) {
            case VALUES:
                tables = new TableValues[limit];
                for (int i = 0; i < limit; i++) {
                    tables[i] = new TableValues(setOfBiclusters[i],
                                                go, mainFrame);
                }
                break;
            case COLORS:
                tables = new TableColors[limit];
                for (int i = 0; i < limit; i++) {
                    tables[i] = new TableColors(setOfBiclusters[i],
                                                go, mainFrame);
                }
                break;
            case SYMBOLS:
                tables = new TableSymbols[limit];
                for (int i = 0; i < limit; i++) {
                    tables[i] = new TableSymbols(setOfBiclusters[i],
                                                 go, mainFrame);
                }
                break;
            }

            /**
             * set the biclusters' number combo box
             */
            JPanel panel = new JPanel();
            BoxLayout bl = new BoxLayout(panel, BoxLayout.X_AXIS);
            panel.setLayout(bl);
            panel.setBorder(BorderFactory.createEmptyBorder(20, 20, 20, 20));
            biclustersNumberOption = new JComboBox(biclustersNumber);
            biclustersNumberOption.setSelectedIndex(selectedIndex);
            panel.add(biclustersNumberOption);
            for (int i = 0; i < 7; i++) {
                panel.add(new JPanel());
            }

            /**
             * add the item listener that will be listening to the changes
             * in items' selection on the combo box
             */
            biclustersNumberOption.addItemListener(new
                    PanelAnalyzingMatrixTables_biclustersNumberOption_itemAdapter(this));
            /**
             * display the combo box
             */
            this.add(panel);
        } catch (InvalidNodeObjectTypeException ex) {
            ex.printStackTrace();
        }

        this.displayTables();
    }

    /**
     * Returns the type of the tables presented in the panel. Table types are:<br/>
     *     <code>PanelAnalyzingMatrixTables.VALUES</code><br/>
     *     <code>PanelAnalyzingMatrixTables.COLORS</code><br/>
     *     <code>PanelAnalyzingMatrixTables.SYMBOLS</code><br/>
     *
     * @return <code>int</code> the panel tables' type
     */
    public int getTableType() {
        return tableType;
    }

    /**
     * Returns the options presented to user about the number of tables
     * that he/she wants to visualize. These options are only displayed
     * to user if this panel holds tables of biclusters' groups.
     *
     * @return the <code>String[]</code> set of options displayed to user
     */
    public static String[] getBiclustersNumber() {
        return biclustersNumber;
    }

    /**
     * Returns the application's main frame.
     *
     * @return the main <code>BiggestsFrame</code>
     */
    public BiggestsFrame getMainFrame() {
        return mainFrame;
    }

    /**
     * Returns the array which contains the tables to be displayed
     * (may contain <code>TableValues</code>s, <code>TableColors</code>s
     * and <code>TableSymbols</code>s, depending on the <code>tableType</code>).
     *
     * @return the set of <code>TableValues</code>s, <code>TableColors</code>s
     *         or <code>TableSymbols</code>s that are being displayed in this panel
     */
    public TableValues[] getTables() {
        return tables;
    }

    /**
     * Returns the array with the scroll panes that hold the tables to be
     * displayed (for <code>VALUES</code> <code>tableType</code> only).
     *
     * @return the set of <code>JScrollPane</code>s used to scroll the
     *         displayed tables
     */
    public JScrollPane[] getScrolls() {
        return scrolls;
    }

    /**
     * Returns the set of panels which holds the tables and their headers,
     * which are displayed in this panel (for <code>COLORS</code>
     * and <code>SYMBOLS</code> <code>tableType</code>s only).
     *
     * @return the set of <code>JPanel</code>s which hold the tables and
     *         their headers
     */
    public JPanel[] getPanels() {
        return panels;
    }

    /**
     * Returns the layout of this panel.
     *
     * @return the <code>BoxLayout</code> of this panel
     */
    public BoxLayout getBoxLayout() {
        return boxLayout;
    }

    /**
     * Returns the set of biclusters from which the set of tables is constructed.
     * This is only true if the object to display is either a group of
     * biclusters or a group of post-processed biclusters.
     *
     * @return the set of <code>Bicluster</code>s to display
     */
    public Bicluster[] getSetOfBiclusters() {
        return setOfBiclusters;
    }

    /**
     * Returns the node information object containing the object with the data
     * to display in the table.
     *
     * @return the <code>NodeInfo</code> object
     */
    public NodeInfo getNodeInfo() {
        return nodeInfo;
    }
}


/**
 * <p>Title: </p>
 *
 * <p>Description: </p>
 */
class PanelAnalyzingMatrixTables_biclustersNumberOption_itemAdapter implements
        ItemListener {
    private PanelAnalyzingMatrixTables adaptee;
    PanelAnalyzingMatrixTables_biclustersNumberOption_itemAdapter(
            PanelAnalyzingMatrixTables adaptee) {
        this.adaptee = adaptee;
    }

    /**
     * Associates the specific handler for the item state event fired by
     * the user's selection on the biclusters number option combo box with
     * the general item state changed event handler.
     *
     * @param e the <code>ItemEvent</code> fired by a user's selection
     *          in a graphic component
     */
    public void itemStateChanged(ItemEvent e) {
        adaptee.biclustersNumberOption_itemStateChanged(e);
    }
}
