package biggests.gui;

import javax.swing.*;

/**
 * <p>Title: Panel with Basic Input/Output Validation</p>
 *
 * <p>Description: Extends <code>JPanel</code> class with some methods to
 *                 perform basic validations over data input by the user.
 *                 Abstract class. Non abstract subclasses must implement
 *                 <code>readParameter(String parameterToRead)</code> method,
 *                 which actually gets the parameter entered by the user as a
 *                 <code>String</code>. All remaining methods of this class
 *                 call <code>readParameter</code> method, convert the
 *                 String to some number format and perform extra validations.</p>
 *
 * <p>Copyright:   Copyright (C) 2007  Joana Gonçalves
 *                 This program is free software; you can redistribute
 *                 it and/or modify it under the terms of the GNU General
 *                 Public License as published by the Free Software
 *                 Foundation; either version 3 of the License, or
 *                 any later version.
 *
 *                 This program is distributed in the hope that it will
 *                 be useful, but WITHOUT ANY WARRANTY; without even the
 *                 implied warranty of MERCHANTABILITY or FITNESS FOR A
 *                 PARTICULAR PURPOSE.  See the
 *                 <a href="http://www.gnu.org/licenses/gpl.html">
 *                 GNU General Public License</a> for more details.
 * </p>
 *
 * @author Joana P. Gonçalves
 */
public abstract class PanelBasicIOValidation extends JPanel {
    /**
     * Abstract method. Should be implemented by non abstract subclasses
     * of this class. Reads a specific parameter entered by the user
     * (the selected location and methodology of reading the parameter
     * is delegated to subclasses, that actually know how it should be done).
     *
     * @param parameterToRead <code>String</code> the name of the parameter
     *        to read (specially important when displaying error messages
     *        to user)
     * @return String the parameter read
     * @throws Exception if it was not possible to read the parameter
     */
    public abstract String readParameter(String parameterToRead) throws
            Exception;

    /**
     * Reads <code>parameterToRead</code> entered by the user and converts
     * it to an integer value. Displays an error message if parameter could
     * not be read or converted to integer. Also throws an exception in
     * these cases.
     *
     * @param parameterToRead <code>String</code> the name of the parameter
     *        to read
     * @return int the parameter read
     * @throws Exception if parameter could not be read or could not be
     *         converted to an integer value
     */
    public int readParameterInteger(String parameterToRead) throws Exception {
        int parameter;
        try {
            parameter = Integer.parseInt(readParameter(parameterToRead));
        } catch (NumberFormatException ex) {
            ex.printStackTrace();
            JOptionPane.showMessageDialog(this, "The " + parameterToRead +
                                          " must be an integer value. ",
                                          parameterToRead +
                                          " must be an integer",
                                          JOptionPane.INFORMATION_MESSAGE);
            throw (new Exception());
        }
        return parameter;
    }

    /**
     * Reads <code>parameterToRead</code> entered by the user, converts
     * it to an integer value and checks if the resulting value is in
     * [<code>min</code>, <code>max</code>] range. Displays an error
     * message if parameter could not be read, could not be converted
     * to integer or if it's under the minimum (<code>min</code>) or
     * over the maximum (<code>max</code>) given thresholds. Also
     * throws an exception in these cases.
     *
     * @param min <code>int</code> minimum threshold
     * @param max <code>int</code> maximum threshold
     * @param parameterToRead <code>String</code> the name of the parameter
     *        to read
     * @return int the parameter read
     * @throws Exception if parameter could not be read, could not be
     *         converted to an integer value or if its under the minimum
     *         or over the maximum value
     */
    public int readParameterInteger(int min, int max, String parameterToRead) throws
            Exception {
        int parameter;
        parameter = readParameterInteger(parameterToRead);
        if ((parameter < min) || (parameter > max)) {
            JOptionPane.showMessageDialog(this, "The " + parameterToRead +
                                          " must be between " + min + " and " +
                                          max,
                                          parameterToRead +
                                          " must be between [" + min + ", " +
                                          max + "]",
                                          JOptionPane.INFORMATION_MESSAGE);
            throw (new Exception());
        }
        return parameter;
    }

    /**
     * Reads <code>parameterToRead</code> entered by the user, converts
     * it to an integer value and checks if the resulting value is above
     * <code>min</code> value. Displays an error message if parameter
     * could not be read, could not be cconverted to integer or if it's
     * under the minimum given threshold (<code>min</code>).
     * Also throws an exception in these cases.
     *
     * @param min <code>int</code> minimum threshold
     * @param parameterToRead <code>String</code> the name of the parameter
     *        to read
     * @return int the parameter read
     * @throws Exception if parameter could not be read, could not be
     *         converted to an integer value or if its under the minimum
     */
    public int readParameterInteger(int min, String parameterToRead) throws
            Exception {
        int parameter;
        parameter = readParameterInteger(parameterToRead);
        if ((parameter < min)) {
            JOptionPane.showMessageDialog(this, "The " + parameterToRead +
                                          " must be equal or greater than " +
                                          min + ".",
                                          parameterToRead +
                                          " must be greater than " + min,
                                          JOptionPane.INFORMATION_MESSAGE);
            throw (new Exception());
        }
        return parameter;
    }

    /**
     * Reads <code>parameterToRead</code> entered by the user and converts
     * it to a floating point (<code>float</code>) value. Displays an
     * error message if parameter could not be read or converted to
     * <code>float</code>.
     * Also throws an exception in these cases.
     *
     * @param parameterToRead <code>String</code> the name of the parameter
     *        to read
     * @return float the parameter read
     * @throws Exception if parameter could not be read or could not be
     *         converted to a <code>float</code> value
     */
    public float readParameterFloat(String parameterToRead) throws Exception {
        float parameter;
        try {
            parameter = Float.parseFloat(readParameter(parameterToRead));
        } catch (NumberFormatException ex) {
            ex.printStackTrace();
            JOptionPane.showMessageDialog(this, "The " + parameterToRead +
                                          " must be a float value. ",
                                          parameterToRead + " must be a float",
                                          JOptionPane.INFORMATION_MESSAGE);
            throw (new Exception());
        }
        return parameter;
    }

    /**
     * Reads <code>parameterToRead</code> entered by the user, converts
     * it to a floating point value (<code>float</code> and checks if
     * the resulting value is in [<code>min</code>, <code>max</code>] range.
     * Displays an error message if parameter could not be read, could not
     * be converted to <code>float</code> or if it's under the minimum
     * (<code>min</code>) or over the maximum (<code>max</code>) given
     * thresholds. Also throws an exception in these cases.
     *
     * @param min <code>float</code> minimum threshold
     * @param max <code>float</code> maximum threshold
     * @param parameterToRead <code>String</code> the name of the parameter
     *        to read
     * @return float the parameter read
     * @throws Exception if parameter could not be read, could not be
     *         converted to a <code>float</code> value or if its under
     *         the minimum or over the maximum value
     */
    public float readParameterFloat(float min, float max,
                                    String parameterToRead) throws Exception {
        float parameter;
        parameter = readParameterFloat(parameterToRead);
        if ((parameter < min) || (parameter > max)) {
            JOptionPane.showMessageDialog(this, "The " + parameterToRead +
                                          " must be between " + min + " and " +
                                          max,
                                          parameterToRead +
                                          " must be between [" + min + ", " +
                                          max + "]",
                                          JOptionPane.INFORMATION_MESSAGE);
            throw (new Exception());
        }
        return parameter;
    }

    /**
     * Reads <code>parameterToRead</code> entered by the user, converts
     * it to a floating point value (<code>float</code>) and checks if
     * the resulting value is above <code>min</code> value. Displays an
     * error message if parameter could not be read, could not be
     * converted to <code>float</code> or if it's under the minimum given
     * threshold (<code>min</code>). Also throws an exception in these cases.
     *
     * @param min <code>float</code> minimum threshold
     * @param parameterToRead <code>String</code> the name of the parameter
     *        to read
     * @return float the parameter read
     * @throws Exception if parameter could not be read, could not be
     *         converted to a <code>float</code> value or if its under
     *         the minimum value (<code>min</code>)
     */

    public float readParameterFloat(float min, String parameterToRead) throws
            Exception {
        float parameter;
        parameter = readParameterFloat(parameterToRead);
        if ((parameter < min)) {
            JOptionPane.showMessageDialog(this, "The " + parameterToRead +
                                          " must be equal or greater than " +
                                          min,
                                          parameterToRead +
                                          " must be greater than " + min,
                                          JOptionPane.INFORMATION_MESSAGE);
            throw (new Exception());
        }
        return parameter;
    }

    /**
     * Reads <code>parameterToRead</code> entered by the user and converts
     * it to a floating point (<code>double</code>) value. Displays an
     * error message if parameter could not be read or converted to
     * <code>double</code>.
     * Also throws an exception in these cases.
     *
     * @param parameterToRead <code>String</code> the name of the parameter
     *        to read
     * @return double the parameter read
     * @throws Exception if parameter could not be read or could not be
     *         converted to a <code>double</code> value
     */
    public double readParameterDouble(String parameterToRead) throws Exception {
        double parameter;
        try {
            parameter = Double.parseDouble(readParameter(parameterToRead));
        } catch (NumberFormatException ex) {
            ex.printStackTrace();
            JOptionPane.showMessageDialog(this, "The " + parameterToRead +
                                          " must be a double value. ",
                                          parameterToRead + " must be a double",
                                          JOptionPane.INFORMATION_MESSAGE);
            throw (new Exception());
        }
        return parameter;
    }

    /**
     * Reads <code>parameterToRead</code> entered by the user, converts
     * it to a floating point value (<code>double</code> and checks if
     * the resulting value is in [<code>min</code>, <code>max</code>] range.
     * Displays an error message if parameter could not be read, could not
     * be converted to <code>double</code> or if it's under the minimum
     * (<code>min</code>) or over the maximum (<code>max</code>) given
     * thresholds. Also throws an exception in these cases.
     *
     * @param min <code>double</code> minimum threshold
     * @param max <code>double</code> maximum threshold
     * @param parameterToRead <code>String</code> the name of the parameter
     *        to read
     * @return double the parameter read
     * @throws Exception if parameter could not be read, could not be
     *         converted to a <code>double</code> value or if its under
     *         the minimum or over the maximum value
     */
    public double readParameterDouble(double min, double max,
                                      String parameterToRead) throws Exception {
        double parameter;
        parameter = readParameterDouble(parameterToRead);
        if ((parameter < min) || (parameter > max)) {
            JOptionPane.showMessageDialog(this, "The " + parameterToRead +
                                          " must be between " + min + " and " +
                                          max,
                                          parameterToRead +
                                          " must be between [" + min + ", " +
                                          max + "]",
                                          JOptionPane.INFORMATION_MESSAGE);
            throw (new Exception());
        }
        return parameter;
    }

    /**
     * Reads <code>parameterToRead</code> entered by the user, converts
     * it to a floating point value (<code>double</code>) and checks if
     * the resulting value is above <code>min</code> value. Displays an
     * error message if parameter could not be read, could not be
     * converted to <code>double</code> or if it's under the minimum given
     * threshold (<code>min</code>). Also throws an exception in these cases.
     *
     * @param min <code>double</code> minimum threshold
     * @param parameterToRead <code>String</code> the name of the parameter
     *        to read
     * @return double the parameter read
     * @throws Exception if parameter could not be read, could not be
     *         converted to a <code>double</code> value or if its under
     *         the minimum value (<code>min</code>)
     */

    public double readParameterDouble(double min, String parameterToRead) throws
            Exception {
        double parameter;
        parameter = readParameterDouble(parameterToRead);
        if ((parameter < min)) {
            JOptionPane.showMessageDialog(this, "The " + parameterToRead +
                                          " must be equal or greater than " +
                                          min,
                                          parameterToRead +
                                          " must be greater than " + min,
                                          JOptionPane.INFORMATION_MESSAGE);
            throw (new Exception());
        }
        return parameter;
    }
}
