package biggests.tables;

import java.util.*;

import javax.swing.table.*;

/**
 * <p>Title: Table Sorter for Tables of Functions of Biclusters' Groups</p>
 *
 * <p>Description: A table model which allows for column sorting.
 *                 This model implements table sorting for tables
 *                 with <code>String</code>s in the first column,
 *                 <code>Integer</code>s in the following two,
 *                 <code>Float</code>s in the third pair of columns
 *                 and <code>Integer</code>s in the last column.
 * </p>
 *
 * <p>Copyright:   Copyright (C) 2007  Joana Gonçalves
 *                 This program is free software; you can redistribute
 *                 it and/or modify it under the terms of the GNU General
 *                 Public License as published by the Free Software
 *                 Foundation; either version 3 of the License, or
 *                 any later version.
 *
 *                 This program is distributed in the hope that it will
 *                 be useful, but WITHOUT ANY WARRANTY; without even the
 *                 implied warranty of MERCHANTABILITY or FITNESS FOR A
 *                 PARTICULAR PURPOSE.  See the
 *                 <a href="http://www.gnu.org/licenses/gpl.html">
 *                 GNU General Public License</a> for more details.
 * </p>
 *
 * @author Joana P. Gonçalves
 */
public class TableSorterFunctionGroup extends TableSorter {
    /**
     * States if this table sorter is for a table of functions of a
     * post-processed biclusters group or not.
     */
    private boolean postProcessed;

    /**
     * Constructs a new <code>TableSorterFunctionGroup</code> model.
     *
     * @param postProcessed <code>boolean</code> <code>true</code> if
     * this table sorter is for a table of functions of a post-processed
     * biclusters group, false otherwise
     * @see TableSorter#TableSorter()
     */
    public TableSorterFunctionGroup(boolean postProcessed) {
        super();
        this.postProcessed = postProcessed;
    }

    /**
     * Constructs a new <code>TableSorterFunctionGroup</code> model.
     *
     * @param tableModel        the <code>TableModel</code> which is the basic
     *                          data model for the table
     * @param postProcessed <code>boolean</code> <code>true</code> if
     * this table sorter is for a table of functions of a post-processed
     * biclusters group, false otherwise
     * @see TableSorter#TableSorter(TableModel tableModel)
     */
    public TableSorterFunctionGroup(TableModel tableModel,
                                    boolean postProcessed) {
        super(tableModel);
        this.postProcessed = postProcessed;
    }

    /**
     * Constructs a new <code>TableSorterFunctionGroup</code> model.
     *
     * @param tableModel        the <code>TableModel</code> which is the basic
     *                          data model for the table
     * @param tableHeader       the <code>TableHeader</code> which is the
     *                          header for the table
     * @param postProcessed <code>boolean</code> <code>true</code> if
     * this table sorter is for a table of functions of a post-processed
     * biclusters group, false otherwise
     * @see TableSorter#TableSorter(TableModel tableModel, JTableHeader tableHeader)
     */
    public TableSorterFunctionGroup(TableModel tableModel,
                                    TableHeader tableHeader,
                                    boolean postProcessed) {
        super(tableModel, tableHeader);
        this.postProcessed = postProcessed;
    }

    /**
     * Overrides <code>TableSorter</code>'s <code>getColumnClass</code> method.
     * Returns the common superclass of the elements at the column with index
     * <code>column</code>.<br/>
     *
     * @param column   <code>int</code> value: the column being queried
     * @return         the common super<code>Class</code> of the column at
     *                 index <code>column</code>
     * @see            TableSorter#getColumnClass(int column)
     */
    public Class<?> getColumnClass(int column) {
        if (postProcessed) {
            if (column == 0 || column == 1 || column == 2 || column == 3 ||
                column == 6 || column == 7 || column == 9) {
                return (Integer.class);
            } else {
                return (Float.class);
            }
        }
        if (column == 0 || column == 1 || column == 2 || column == 5 ||
            column == 6 || column == 8) {
            return (Integer.class);
        } else {
            return (Float.class);
        }
    }

    /**
     * Overrides <code>TableSorter</code>'s <code>getComparator</code> method.
     * Returns the suitable comparator for the values of column number
     * <code>column</code>.
     *
     * @param column the <code>int</code> value which is the index of the column
     * @return the suitable <code>Comparator</code> for column <code>column</code>
     * @see TableSorter#getComparator(int column)
     */
    protected Comparator getComparator(int column) {
        Class columnType = getColumnClass(column);
        Comparator comparator = (Comparator) columnComparators.get(columnType);

        if (comparator != null) {
            return comparator;
        }
        if (Comparable.class.isAssignableFrom(columnType)) {
            return COMPARABLE_COMPARATOR;
        }
        return LEXICAL_COMPARATOR;
    }
}
