package biggests.tables;

import java.awt.*;
import java.awt.event.*;
import java.text.*;
import java.util.*;

import javax.swing.*;
import javax.swing.table.*;

import biggests.utils.*;
import smadeira.biclustering.*;
import smadeira.ontologizer.*;
import ontologizer.go.TermID;
import ontologizer.ByteString;
import ontologizer.association.Gene2Associations;

/**
 * <p>Title: Table of Values</p>
 *
 * <p>Description:
 *                 <p>Defines a table of values with the following data:<br/>
 *                 - header -> conditions' names (more widely, <code>String</code>s)<br/>
 *                 - first column -> genes' names (more widely, <code>String</code>s)<br/>
 *                 - all other columns' cells -> gene expression values
 *                 (more widely, <code>float</code> values)</p>
 *
 *                 <p>This is a simple table that shows the raw values that
 *                 are assigned when using one of the constructors and
 *                 implements sorting functionality on the columns, i.e.,
 *                 allows the reordering of the rows, sorting them by a
 *                 specific column's cells values in ascending or descending
 *                 order.</p>
 *                 </p>
 *
 * <p>Copyright:   Copyright (C) 2007  Joana Gonçalves
 *                 This program is free software; you can redistribute
 *                 it and/or modify it under the terms of the GNU General
 *                 Public License as published by the Free Software
 *                 Foundation; either version 3 of the License, or
 *                 any later version.
 *
 *                 This program is distributed in the hope that it will
 *                 be useful, but WITHOUT ANY WARRANTY; without even the
 *                 implied warranty of MERCHANTABILITY or FITNESS FOR A
 *                 PARTICULAR PURPOSE.  See the
 *                 <a href="http://www.gnu.org/licenses/gpl.html">
 *                 GNU General Public License</a> for more details.
 * </p>
 *
 * @author Joana P. Gonçalves
 */

public class TableValues extends JTable {
    /**
     *
     */
    protected GOFrontEnd goFrontEnd;

    protected Component parent;

    /**
     * Creates a new values table from the data of a node object. A table of
     * values contains the following data:<br/>
     * - header -> conditions' names<br/>
     * - first column -> genes' names<br/>
     * - all other columns' cells -> gene expression float values<br/>
     * This table also displays the terms annotated with each gene
     * (when a specific table's row is clicked by the user).
     *
     * @param nodeObject the <code>NodeObjectInterface</code> from which the
     *                   values table is created
     * @param go the <code>GOFrontEnd</code>
     * @param parent the GUI's parent <code>Component</code>
     * @throws InvalidNodeObjectTypeException if the type of the node
     *                   object is not valid
     * @see #constructRowData(NodeObjectInterface nodeObject)
     * @see #constructColumnNames(NodeObjectInterface nodeObject)
     * @see #TableValues(Object[][] rowData, Object[] columnNames)
     */
    /* ---------------------------------------------------------------------<br/>
     * | Genes Names | Condition1 | Condition2 | Condition3 | .............|<br/>
     * |-------------------------------------------------------------------|<br/>
     * | Gene1       | 0.26394536 | .......... | .......... | ............ |<br/>
     * | Gene2       | -0.0107532 | .......... | .......... | ............ |<br/>
     * | Gene3       | .......... | .......... | .......... | ............ |<br/>
     * | Gene4       | .......... | .......... | .......... | ............ |<br/>
     * ---------------------------------------------------------------------
     */
    public TableValues(NodeObjectInterface nodeObject, GOFrontEnd go,
                       Component parent) throws InvalidNodeObjectTypeException {
        this(constructRowData(nodeObject), constructColumnNames(nodeObject));
        this.goFrontEnd = go;
        this.parent = parent;
    }


    /**
     * Creates a new values table from the data of a node object. A table of
     * values contains the following data:<br/>
     * - header -> conditions' names<br/>
     * - first column -> genes' names<br/>
     * - all other columns' cells -> gene expression float values<br/>
     *
     * @param nodeObject the <code>NodeObjectInterface</code> from which the
     *                   values table is created
     * @throws InvalidNodeObjectTypeException when the type of the node
     *                   object is not valid
     * @see #constructRowData(NodeObjectInterface nodeObject)
     * @see #constructColumnNames(NodeObjectInterface nodeObject)
     * @see #TableValues(Object[][] rowData, Object[] columnNames)
     */
    /* ---------------------------------------------------------------------<br/>
     * | Genes Names | Condition1 | Condition2 | Condition3 | .............|<br/>
     * |-------------------------------------------------------------------|<br/>
     * | Gene1       | 0.26394536 | .......... | .......... | ............ |<br/>
     * | Gene2       | -0.0107532 | .......... | .......... | ............ |<br/>
     * | Gene3       | .......... | .......... | .......... | ............ |<br/>
     * | Gene4       | .......... | .......... | .......... | ............ |<br/>
     * ---------------------------------------------------------------------
     */
    public TableValues(NodeObjectInterface nodeObject) throws
            InvalidNodeObjectTypeException {
        this(constructRowData(nodeObject), constructColumnNames(nodeObject));
    }

    /**
     * Creates a new values table with the given row data and column names. A
     * table of values contains the following data:<br/>
     * - header -> conditions' names (more widely, <code>String</code>s)<br/>
     * - first column -> genes' names (more widely, <code>String</code>s)<br/>
     * - all other columns' cells -> gene expression values (more widely,
     *   <code>float</code> values)<br/>
     *
     * @param rowData the <code>Object[][]</code> row data (for all table
     *                cells except the ones from the table header)
     * @param columnNames the <code>Object[]</code> column names (for
     *                the table header)
     * @see #TableValues(TableSorterValues tableSorterModel)
     * @see biggests.tables.TableSorter#TableSorter(TableModel tableModel)
     * @see javax.swing.table.DefaultTableModel#DefaultTableModel(Object[][] rowData, Object[] columnNames)
     */
    /* ---------------------------------------------------------------------<br/>
     * | Genes Names | Condition1 | Condition2 | Condition3 | .............|<br/>
     * |-------------------------------------------------------------------|<br/>
     * | Gene1       | 0.26394536 | .......... | .......... | ............ |<br/>
     * | Gene2       | -0.0107532 | .......... | .......... | ............ |<br/>
     * | Gene3       | .......... | .......... | .......... | ............ |<br/>
     * | Gene4       | .......... | .......... | .......... | ............ |<br/>
     * ---------------------------------------------------------------------
     */
    protected TableValues(Object[][] rowData, Object[] columnNames) {
        /**
         * create a new JTable with a table sorter model, which incorporates
         * the specified row data and column names
         */
        this(new TableSorterValues(new DefaultTableModel(rowData, columnNames)));
    }

    /**
     * Creates a new values table with the given row data and column names. A
     * table of values contains the following data:<br/>
     * - header -> conditions' names (more widely, <code>String</code>s)<br/>
     * - first column -> genes' names (more widely, <code>String</code>s)<br/>
     * - all other columns' cells -> gene expression values (more widely,
     *   <code>float</code> values)<br/>
     *
     * @param tableSorterModel the <code>TableSorter</code> model for the table
     * @see javax.swing.JTable#JTable(TableModel dm)
     */
    /* ---------------------------------------------------------------------<br/>
     * | Genes Names | Condition1 | Condition2 | Condition3 | .............|<br/>
     * |-------------------------------------------------------------------|<br/>
     * | Gene1       | 0.26394536 | .......... | .......... | ............ |<br/>
     * | Gene2       | -0.0107532 | .......... | .......... | ............ |<br/>
     * | Gene3       | .......... | .......... | .......... | ............ |<br/>
     * | Gene4       | .......... | .......... | .......... | ............ |<br/>
     * ---------------------------------------------------------------------
     */
    protected TableValues(TableSorterValues tableSorterModel) {
        /**
         * create a new JTable object with a table sorter model derived from the
         * given table model
         */
        super(tableSorterModel);
        this.addMouseListener(new TableValues_this_mouseAdapter(this));

        /**
         * set table header of this TableValues as the table header of the
         * created table sorter model
         */
        ((TableSorterValues)this.getModel()).setTableHeader(this.getTableHeader());

        this.getTableHeader().setReorderingAllowed(false);

        this.setDefaultRenderer(Float.class, new FloatCellRenderer());

        /**
         * set autoresize (of columns' width) off
         */
        this.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);

        /**
         * set preferred size for all table's cells
         */
        Enumeration<TableColumn> e = this.getColumnModel().getColumns();
        while (e.hasMoreElements()) {
            e.nextElement().setPreferredWidth(100);
        }

        /**
         * set header font to bold style
         */
        JTableHeader tableHeader = this.getTableHeader();
        Font headerFont = tableHeader.getFont();
        tableHeader.setFont(new Font(headerFont.getName(), Font.BOLD,
                                     headerFont.getSize()));
    }

    /**
     * Overrides <code>JTable</code>'s <code>prepareRenderer</code> method.
     * Prepares the renderer by querying the data model for the
     * value and selection state
     * of the cell at <code>row</code>, <code>column</code>.
     * Returns the component (may be a <code>Component</code>
     * or a <code>JComponent</code>) under the event location.
     *
     * @param renderer  the <code>TableCellRenderer</code> to prepare
     * @param row       the row of the cell to render, where 0 is the first row
     * @param column    the column of the cell to render,
     *			where 0 is the first column
     * @return          the <code>Component</code> under the event location
     * @see javax.swing.JTable#prepareRenderer(TableCellRenderer renderer, int row, int column)
     */
    public Component prepareRenderer(TableCellRenderer renderer, int row,
                                     int column) {
        /**
         * get the component prepared from JTable's prepareRenderer method
         */
        Component c = super.prepareRenderer(renderer, row, column);

        if ((row % 2 == 0) && (!isCellSelected(row, column))) {
            /**
             * cell is in row numbered even and is not selected at the moment
             * set the color of the cell to light blue
             */
            c.setBackground(new Color(228, 238, 250));
        } else
        if (!isCellSelected(row, column)) {
            /**
             * cell is in a row numbered odd and is not selected at the moment
             * set the color of the cell to default background
             */
            c.setBackground(getBackground());
        }
        return c;
    }

    /**
     * Overrides <code>JTable</code>'s <code>getToolTipText</code> method
     * in order to allow the renderer's tips to be used if they have text set.
     *
     * @param event MouseEvent the event that triggered the tool tip display
     * @return String the message set as tool tip for the pointed table cell
     * @see javax.swing.JTable#getToolTipText(MouseEvent event)
     */
    public String getToolTipText(MouseEvent event) {
        String tip = null;
        Point p = event.getPoint();
        int rowIndex = rowAtPoint(p);
        int colIndex = columnAtPoint(p);
        int realColumnIndex = convertColumnIndexToModel(colIndex);

        if (rowIndex < 0) {
            return "";
        }

        if (realColumnIndex == 0) {
            /**
             * cell is in the first column, so it contains the name of a gene
             * set the tex tip for this cell, something like: "Gene X"
             */
            tip = "Gene: " + getValueAt(rowIndex, 0);
        } else
        if ((rowIndex >= 0) && (rowIndex < this.getRowCount()) &&
            (realColumnIndex > 0) && (realColumnIndex < this.getColumnCount())) {
            /**
             * cell is in a column that is not the first one, so it contains
             * a float value (the gene expression value);
             * set the tip for this cell, something like:
             * "(gene_name, condition_name) = gene_expression_value"
             */
            tip = "(" + getValueAt(rowIndex, 0) +
                  ", " + getColumnName(realColumnIndex) +
                  ") = " + getValueAt(rowIndex, realColumnIndex);
        } else {
            tip = super.getToolTipText(event);
        }
        return tip;
    }

    /**
     * Overrides <code>JTable</code>'s <code>createDefaultTableHeader</code>
     * method.<br/>
     * Creates the default table header for a values' table, that is, a header
     * with specific text tool tips for this kind of table.
     *
     * @return an instance of <code>TableHeader</code>, i.e, a specific table
     *         header for a table of values
     * @see javax.swing.JTable#columnModel
     * @see javax.swing.JTable#createDefaultTableHeader()
     * @see biggests.tables.TableHeader#TableHeader(TableColumnModel cm)
     */
    protected JTableHeader createDefaultTableHeader() {
        return new TableHeader(columnModel);
    }

    /**
     * Constructs the row data matrix with data retrieved from the node object.<br/>
     * <p>
     * The final structure of the row data is:<br/>
     * - the first column corresponds to the names of the genes<br/>
     * - the other cells of the matrix contain float values, which are the gene
     *   expression values under certain conditions
     * </p>
     * <p>
     * <code>protected</code>: it's intended to be used only by the
     *                       <code>TableValues</code> constructor and
     *                       constructors of <code>TableValues</code>'s
     *                       subclasses.<br/>
     * <code>static</code>: because object <code>this</code> cannot be
     *                      referrenced by the constructor before it is
     *                      actually created.
     * </p>
     *
     * @param nodeObject the <code>NodeObjectInterface</code> which provides the
     *                   data needed to create the row data matrix
     * @return the <code>Object[][]</code> row data matrix, ready to be used by
     *         <code>TableValues</code>'s constructor
     * @throws InvalidNodeObjectTypeException when the type of the node object is not valid
     */
    protected static Object[][] constructRowData(NodeObjectInterface nodeObject) throws
            InvalidNodeObjectTypeException {
        String[] geneNames = null;
        String[] conditionNames = null;
        float[][] valuesMatrix = null;
        Object[][] rowData = null;

        if (nodeObject instanceof ExpressionMatrix) {
            /**
             * node object is an expression or preprocessed expression matrix
             */
            ExpressionMatrix expressionMatrix = (ExpressionMatrix) nodeObject;

            /**
             * get genes and condition names for this matrix
             */
            geneNames = expressionMatrix.getGenesNames();
            conditionNames = expressionMatrix.getConditionsNames();

            /**
             * get the matrix itself, with the gene expression values
             */
            valuesMatrix = expressionMatrix.getGeneExpressionMatrix();
        } else
        if (nodeObject instanceof PreProcessedExpressionMatrix) {
            /**
             * node object is an expression or preprocessed expression matrix
             */
            PreProcessedExpressionMatrix preProcessedExpressionMatrix = (
                    PreProcessedExpressionMatrix) nodeObject;

            /**
             * get genes and condition names for this matrix
             */
            geneNames = preProcessedExpressionMatrix.getGenesNames();
            conditionNames = preProcessedExpressionMatrix.getConditionsNames();

            /**
             * get the matrix itself, with the gene expression values
             */
            valuesMatrix = preProcessedExpressionMatrix.getGeneExpressionMatrix();
        } else
        if (nodeObject instanceof DiscretizedExpressionMatrix) {
            /**
             * node object is a discretized expression matrix
             */
            AbstractExpressionMatrix originalExpressionMatrix =
                    ((DiscretizedExpressionMatrix) nodeObject).
                    getOriginalExpressionMatrix();

            /**
             * get gene and condition names
             */
            geneNames = originalExpressionMatrix.getGenesNames();
            conditionNames = originalExpressionMatrix.getConditionsNames();

            /**
             * get the expression or preprocessed matrix from which this
             * discretized matrix was constructed
             */
            valuesMatrix = originalExpressionMatrix.getGeneExpressionMatrix();
        } else
        if ((nodeObject instanceof Bicluster) ||
            (nodeObject instanceof CCC_Bicluster) ||
            (nodeObject instanceof E_CCC_Bicluster)) {
            /**
             * node object is a bicluster or a post-processed bicluster
             */
            Bicluster bicluster = (Bicluster) nodeObject;

            /**
             * get gene and condition names
             */
            geneNames = bicluster.getGenesNames();
            conditionNames = bicluster.getConditionsNames();

            /**
             * get the bicluster/post-processed bicluster expression matrix
             */
            valuesMatrix = bicluster.getBiclusterExpressionMatrix();
        } else {
            throw (new InvalidNodeObjectTypeException("Only original, " +
                    "preprocessed, discretized matrices and individual " +
                    "biclusters are allowed."));
        }

        /**
         * construct row data matrix:
         * - first column cells have the gene names
         * - all other column cells contain gene expression values
         */
        rowData = new Object[geneNames.length][conditionNames.length + 1];
        for (int i = 0; i < geneNames.length; i++) {
            rowData[i][0] = geneNames[i];
            for (int j = 1; j < conditionNames.length + 1; j++) {
                rowData[i][j] = Float.valueOf(valuesMatrix[i][j - 1]);
            }
        }
        return rowData;
    }

    /**
     * Constructs the column names' array with data retrieved from the node object.<br/>
     * <p>
     * The final structure of the column names' array is:<br/>
     * - the first cell contains the gene column name<br/>
     * - the remaining cells contain the condition names.
     * </p>
     * <p>
     * <code>protected</code>: it's intended to be used only by the
     *                       <code>TableValues</code>'s constructor and
     *                       <code>TableValues</code> subclasses'
     *                       constructors.<br/>
     * <code>static</code>: because object <code>this</code> cannot be
     *                      referrenced by the constructor before it is
     *                      actually created.
     * </p>
     *
     * @param nodeObject the <code>NodeObjectInterface</code> which provides the
     *                   data needed to create the column names' array
     * @return the <code>Object[]</code> array with the column names
     * @throws InvalidNodeObjectTypeException when the type of the node object is not valid
     */
    protected static Object[] constructColumnNames(NodeObjectInterface
            nodeObject) throws InvalidNodeObjectTypeException {
        String[] conditionNames = null;

        if (nodeObject instanceof ExpressionMatrix) {
            /**
             * node object is an expression or preprocessed expression matrix
             */
            ExpressionMatrix expressionMatrix = (ExpressionMatrix) nodeObject;

            /**
             * get the names of the conditions
             */
            conditionNames = expressionMatrix.getConditionsNames();
        } else
        if (nodeObject instanceof PreProcessedExpressionMatrix) {
            /**
             * node object is an expression or preprocessed expression matrix
             */
            PreProcessedExpressionMatrix preProcessedExpressionMatrix = (
                    PreProcessedExpressionMatrix) nodeObject;

            /**
             * get the names of the conditions
             */
            conditionNames = preProcessedExpressionMatrix.getConditionsNames();
        } else
        if (nodeObject instanceof DiscretizedExpressionMatrix) {
            /**
             * node object is a discretized expression matrix
             */
            DiscretizedExpressionMatrix discretizedExpressionMatrix = (
                    DiscretizedExpressionMatrix) nodeObject;

            /**
             * get the conditions' names of the expression (or preprocessed
             * expression) matrix from which the discretized matrix was derived
             */
            conditionNames = discretizedExpressionMatrix.
                             getOriginalExpressionMatrix().getConditionsNames();
        } else
        if ((nodeObject instanceof Bicluster) ||
            (nodeObject instanceof CCC_Bicluster) ||
            (nodeObject instanceof E_CCC_Bicluster)) {
            /**
             * node object is a bicluster (or post-processed bicluster)
             */
            Bicluster bicluster = (Bicluster) nodeObject;

            /**
             * get the names of the conditions in the bicluster
             */
            conditionNames = bicluster.getConditionsNames();
        } else {
            throw (new InvalidNodeObjectTypeException("Only original, " +
                    "preprocessed, discretized matrices and individual " +
                    "biclusters are allowed."));
        }

        /**
         * construct the column names for this table:
         * - first cell contains a label for the first column (the column which
         * holds the genes' names) - "Genes Names"
         * - each of the other cells contain the name of the corresponding
         * condition
         */
        String[] columnNames = new String[conditionNames.length + 1];

        columnNames[0] = "Genes Names";
        for (int i = 1; i < columnNames.length; i++) {
            columnNames[i] = conditionNames[i - 1];
        }
        return columnNames;
    }

    public void this_mouseClicked(MouseEvent e) {
        if (e.getButton() == MouseEvent.BUTTON1) {
            if (goFrontEnd == null /*|| goProperties == null*/) {
                return;
            }

            Point p = e.getPoint();
            int viewRowIndex = this.rowAtPoint(p);

            if (viewRowIndex < 0) return;

            String gene = (String)this.getValueAt(viewRowIndex, 0);
            String goTermIDsAndDescriptions = "";

/***************************************************************************
* THE COUNT IS THE NUMBER OF DIRECT ASSOCIATIONS
*/

            Gene2Associations gene2Associations =
                    goFrontEnd.getGoAssociations().get(new ByteString(gene));
            if (gene2Associations == null) {
                JOptionPane.showMessageDialog(
                        parent,
                        "Gene ID or ORF was not found." +
                        gene,
                        "Gene was not found",
                        JOptionPane.INFORMATION_MESSAGE);
                return;
            }

            ArrayList<TermID> associations = gene2Associations.getAssociations();
            int count = associations.size();
            for (int i = 0; i < associations.size(); i++) {
                goTermIDsAndDescriptions += "\n" +
                        associations.get(i) + "\t" +
                        goFrontEnd.getGoTerms().getGOName(associations.get(i));
            }

            JTextArea textArea;
            if (count == 0) {
                JOptionPane.showMessageDialog(parent,
                                              "No GO terms annotated with gene " +
                                              gene,
                                              "No GO terms annotated",
                                              JOptionPane.INFORMATION_MESSAGE);
                return;
            }
            textArea = new JTextArea("Gene: " + gene +
                                     "\nAnnotated GO Terms: " + count +
                                     "\n" + goTermIDsAndDescriptions);
            textArea.setEditable(false);
            JScrollPane scroll = new JScrollPane(textArea);
            scroll.setPreferredSize(new Dimension(450, 250));
            JOptionPane.showMessageDialog(parent, scroll, "Annotated GO Terms",
                                          JOptionPane.INFORMATION_MESSAGE);


            /*************************************************************************
             /* THE COUNT IS PERFORMED APPLYING THE TRUE PATH RULE
                          int count = 0;
            for (int i = 0; i < goProperties.getNumberOfTerms(); i++) {
                if (goFrontEnd.termQueryFull(Integer.parseInt(goProperties.
                        getTermID()[i].substring(3)), gene)) {
                    goTermIDsAndDescriptions += "\n" +
                            goProperties.getTermID()[i] + "\t" +
                            goProperties.getTermName()[i];
                    count++;
                }
            }
 */
        }
    }

    /**
     *
     * <p>Title: <code>NumberCellRenderer</code>: A renderer for cells which
     *           contain numbers.</p>
     *
     * <p>Description: </p>
     */
    protected static class NumberCellRenderer extends DefaultTableCellRenderer.
            UIResource {
        /**
         * Constructs a new <code>NumberCellRenderer</code> for a cell.
         *
         * @see javax.swing.table.DefaultTableCellRenderer#DefaultTableCellRenderer()
         */
        public NumberCellRenderer() {
            super();
            setHorizontalAlignment(JLabel.RIGHT);
        }
    }


    /**
     * <p>Title: <code>FloatCellRenderer</code>: A renderer for cells which
     *           contain float numbers.</p>
     *
     * <p>Description: </p>
     */
    protected static class FloatCellRenderer extends NumberCellRenderer {
        /**
         * The formatter which is used to format the float value.
         */
        NumberFormat formatter;

        /**
         * Constructs a new <code>FloatCellRenderer</code>.
         */
        public FloatCellRenderer() {
            super();
        }

        /**
         * Overrides <code>DefaultTableCellRenderer</code>'s
         * <code>setValue</code> method. Formats the value of the cell to display.
         *
         * @param value the <code>Object</code> containing the float value
         */
        public void setValue(Object value) {
            if (formatter == null) {
                formatter = NumberFormat.getInstance(Locale.ENGLISH);
                formatter.setMaximumFractionDigits(8);
                formatter.setMinimumFractionDigits(8);
            }
            setText((value == null) ? "" : formatter.format(value));
        }
    }
}


class TableValues_this_mouseAdapter extends MouseAdapter {
    private TableValues adaptee;
    TableValues_this_mouseAdapter(TableValues adaptee) {
        this.adaptee = adaptee;
    }

    public void mouseClicked(MouseEvent e) {
        adaptee.this_mouseClicked(e);
    }
}
